'''
Module to store conda build settings.
'''
from __future__ import absolute_import, division, print_function

import copy
from collections import namedtuple
import math
import os
from os.path import abspath, expanduser, join, expandvars
import shutil
import sys
import time

from .conda_interface import root_dir, root_writable
from .conda_interface import binstar_upload
from .variants import get_default_variant
from .conda_interface import cc_platform, cc_conda_build, subdir

from .utils import get_build_folders, rm_rf, get_logger, get_conda_operation_locks

on_win = (sys.platform == 'win32')

# Don't "save" an attribute of this module for later, like build_prefix =
# conda_build.config.config.build_prefix, as that won't reflect any mutated
# changes.

conda_build = "conda-build"


def _ensure_dir(path):
    # this can fail in parallel operation, depending on timing.  Just try to make the dir,
    #    but don't bail if fail.
    if not os.path.isdir(path):
        try:
            os.makedirs(path)
        except OSError:
            pass


# we need this to be accessible to the CLI, so it needs to be more static.
DEFAULT_PREFIX_LENGTH = 255

# translate our internal more meaningful subdirs to the ones that conda understands
SUBDIR_ALIASES = {
    'linux-cos5-x86_64': 'linux-64',
    'linux-cos6-x86_64': 'linux-64',
    'linux-cos5-x86': 'linux-32',
    'linux-cos6-x86': 'linux-32',
    'osx-109-x86_64': 'osx-64',
    'win-x86_64': 'win-64',
    'win-x86': 'win-32',
}


Setting = namedtuple("ConfigSetting", "name, default")
DEFAULTS = [Setting('activate', True),
            Setting('anaconda_upload', binstar_upload),
            Setting('force_upload', True),
            Setting('channel_urls', []),
            Setting('dirty', False),
            Setting('include_recipe', True),
            Setting('no_download_source', False),
            Setting('override_channels', False),
            Setting('skip_existing', False),
            Setting('token', None),
            Setting('user', None),
            Setting('verbose', True),

            Setting('debug', False),
            Setting('timeout', 90),
            Setting('set_build_id', True),
            Setting('disable_pip', False),
            Setting('_output_folder', None),
            Setting('prefix_length_fallback', True),
            Setting('_prefix_length', DEFAULT_PREFIX_LENGTH),
            Setting('long_test_prefix', True),
            Setting('locking', True),
            Setting('max_env_retry', 3),
            Setting('remove_work_dir', True),
            Setting('_host_platform', None),
            Setting('_host_arch', None),
            Setting('filename_hashing', cc_conda_build.get('filename_hashing',
                                                           'true').lower() == 'true'),
            Setting('keep_old_work', False),
            Setting('_src_cache_root', abspath(expanduser(expandvars(
                cc_conda_build.get('cache_dir')))) if cc_conda_build.get('cache_dir') else None),
            Setting('copy_test_source_files', True),

            # should rendering cut out any skipped metadata?
            Setting('trim_skip', True),

            # Disable the overlinking test for this package. This test checks that transitive DSOs
            # are not referenced by DSOs in the package being built. When this happens something
            # has gone wrong with:
            # 1. Linker flags not being passed, or not working correctly:
            #    (GNU ld: -as-needed, Apple ld64: -dead_strip_dylibs -no_implicit_dylibs)
            # 2. A missing package in reqs/run (maybe that package is missing run_exports?)
            # 3. A missing (or broken) CDT package in reqs/build or (on systems without CDTs)
            # 4. .. a missing value in the hard-coded but metadata-augmentable library whitelist
            # It is important that packages do not suffer from 2 because uninstalling that missing
            # package leads to an inability to run this package.
            #
            # default to not erroring with overlinking for now.  We have specified in
            #    cli/main_build.py that this default will switch in conda-build 4.0.
            Setting('error_overlinking', cc_conda_build.get('error_overlinking',
                                                           'false').lower() == 'true'),

            Setting('index', None),
            # support legacy recipes where only build is specified and expected to be the
            #    folder that packaging is done on
            Setting('build_is_host', False),

            # these are primarily for testing.  They override the native build platform/arch,
            #     which is useful in tests, but makes little sense on actual systems.
            Setting('_platform', None),
            Setting('_arch', None),
            Setting('_target_subdir', None),

            # variants
            Setting('variant_config_files', []),
            # this file precludes usage of any system-wide or cwd config files.  Config files in
            #    recipes are still respected, and they override this file.
            Setting('exclusive_config_file', None),
            Setting('ignore_system_variants', False),
            Setting('hash_length', 7),

            # append/clobber metadata section data (for global usage.  Can also add files to
            #    recipe.)
            Setting('append_sections_file', None),
            Setting('clobber_sections_file', None),
            Setting('bootstrap', None),

            # source provisioning.
            Setting('git_commits_since_tag', 0),

            # pypi upload settings (twine)
            Setting('password', None),
            Setting('sign', False),
            Setting('sign_with', 'gpg'),
            Setting('identity', None),
            Setting('config_file', None),
            Setting('repository', 'pypitest'),

            # conda-verify is disabled by default.  This is because it was inoperable for a long
            #     time, and thus needs to be opt-in not opt-out.
            Setting('verify', False),
            Setting('ignore_recipe_verify_scripts',
                  cc_conda_build.get('ignore_recipe_verify_scripts', [])),
            Setting('ignore_package_verify_scripts',
                    cc_conda_build.get('ignore_package_verify_scripts', [])),
            Setting('run_recipe_verify_scripts',
                    cc_conda_build.get('run_package_verify_scripts', [])),
            Setting('run_package_verify_scripts',
                    cc_conda_build.get('run_package_verify_scripts', [])),

            # Recipes that have no host section, only build, should bypass the build/host line.
            # This is to make older recipes still work with cross-compiling.  True cross-compiling
            # involving compilers (not just python) will still require recipe modification to have
            # distinct host and build sections, but simple python stuff should work without.
            Setting('merge_build_host', False),

            # path to output build statistics to
            Setting('stats_file', None),

            # extra deps to add to test env creation
            Setting('extra_deps', [])
            ]


def print_function_deprecation_warning(func):
    def func_wrapper(*args, **kw):
        log = get_logger(__name__)
        log.warn("WARNING: attribute {} is deprecated and will be removed in conda-build 4.0.  "
                "Please update your code - file issues on the conda-build issue tracker "
                 "if you need help.".format(func.__name__))
        return func(*args, **kw)
    return func_wrapper


class Config(object):
    __file__ = __path__ = __file__
    __package__ = __package__
    __doc__ = __doc__

    def __init__(self, variant=None, **kwargs):
        super(Config, self).__init__()
        # default variant is set in render's distribute_variants
        self.variant = variant or {}
        self.set_keys(**kwargs)
        if self._src_cache_root:
            self._src_cache_root = os.path.expanduser(self._src_cache_root)

    def _set_attribute_from_kwargs(self, kwargs, attr, default):
        value = kwargs.get(attr, getattr(self, attr) if hasattr(self, attr) else default)
        setattr(self, attr, value)
        if attr in kwargs:
            del kwargs[attr]

    def set_keys(self, **kwargs):
        def env(lang, default):
            version = kwargs.pop(lang, None)
            if not version:
                # Hooray for corner cases.
                if lang == 'python':
                    lang = 'py'
                elif lang == 'numpy':
                    lang = 'npy'
                elif lang == 'r_base':
                    lang = 'r'
                var = 'CONDA_' + lang.upper()
                version = os.getenv(var) if os.getenv(var) else default
            elif isinstance(version, list) and len(version) == 1:
                version = version[0]
            return version

        def set_lang(variant, lang):
            value = env(lang, self.variant.get(lang))
            if value:
                if '.' not in str(value):
                    value = '.'.join((value[0], value[1:]))
                variant[lang] = value

        # this is where we override any variant config files with the legacy CONDA_* vars
        #     or CLI params
        for lang in ('perl', 'lua', 'python', 'numpy', 'r_base'):
            set_lang(self.variant, lang)

        self._build_id = kwargs.pop('build_id', getattr(self, '_build_id', ""))
        source_cache = kwargs.pop('cache_dir', None)
        croot = kwargs.pop('croot', None)

        if source_cache:
            self._src_cache_root = os.path.abspath(os.path.normpath(
                os.path.expanduser(source_cache)))
        if croot:
            self._croot = os.path.abspath(os.path.normpath(croot))
        else:
            # set default value (not actually None)
            self._croot = getattr(self, '_croot', None)

        # handle known values better than unknown (allow defaults)
        for value in DEFAULTS:
            self._set_attribute_from_kwargs(kwargs, value.name, value.default)

        # dangle remaining keyword arguments as attributes on this class
        for name, value in kwargs.items():
            setattr(self, name, value)

    @property
    def arch(self):
        """Always the native (build system) arch, except when pretending to be some
        other platform"""
        return self._arch or subdir.rsplit('-', 1)[1]

    @arch.setter
    def arch(self, value):
        log = get_logger(__name__)
        log.warn("setting build arch.  This is only useful when pretending to be on another "
                 "arch, such as for rendering necessary dependencies on a non-native arch."
                 "  I trust that you know what you're doing.")
        self._arch = str(value)

    @property
    def platform(self):
        """Always the native (build system) OS, except when pretending to be some
        other platform"""
        return self._platform or subdir.rsplit('-', 1)[0]

    @platform.setter
    def platform(self, value):
        log = get_logger(__name__)
        log.warn("setting build platform. This is only useful when "
                "pretending to be on another " "another platform, such as "
                "for rendering necessary dependencies on a non-native "
                "platform. I trust that you know what you're doing.")
        if value == 'noarch':
            raise ValueError("config platform should never be noarch.  Set host_platform instead.")
        self._platform = value

    @property
    def build_subdir(self):
        """Determines channel to download build env packages from.
        Should generally be the native platform.  Does not preclude packages from noarch."""
        return '-'.join((self.platform, self.arch))

    @property
    def host_arch(self):
        return (self._host_arch or
                self.variant.get('target_platform', self.build_subdir).split('-', 1)[1])

    @host_arch.setter
    def host_arch(self, value):
        self._host_arch = value

    @property
    def noarch(self):
        return self.host_platform == 'noarch'

    def reset_platform(self):
        if not self.platform == cc_platform:
            self.platform = cc_platform

    @property
    def subdir(self):
        return "-".join([self.platform, str(self.arch)])

    @property
    def host_platform(self):
        return (self._host_platform or
                self.variant.get('target_platform', self.build_subdir).split('-', 1)[0])

    @host_platform.setter
    def host_platform(self, value):
        self._host_platform = value

    @property
    def host_subdir(self):
        subdir = self.variant.get('target_platform', self.build_subdir)
        if self.host_platform == 'noarch':
            subdir = self.host_platform
        elif subdir != "-".join([self.host_platform, str(self.host_arch)]):
            subdir = "-".join([self.host_platform, str(self.host_arch)])
        return SUBDIR_ALIASES.get(subdir, subdir)

    @host_subdir.setter
    def host_subdir(self, value):
        value = SUBDIR_ALIASES.get(value, value)
        values = value.rsplit('-', 1)
        self.host_platform = values[0]
        if len(values) > 1:
            self.host_arch = values[1]

    @property
    def target_subdir(self):
        return self._target_subdir or self.host_subdir

    @target_subdir.setter
    def target_subdir(self, value):
        self._target_subdir = value

    @property
    def src_cache_root(self):
        return self._src_cache_root if self._src_cache_root else self.croot

    @src_cache_root.setter
    def src_cache_root(self, value):
        self._src_cache_root = value

    @property
    def croot(self):
        """This is where source caches and work folders live"""
        if not self._croot:
            _bld_root_env = os.getenv('CONDA_BLD_PATH')
            _bld_root_rc = cc_conda_build.get('root-dir')
            if _bld_root_env:
                self._croot = abspath(expanduser(_bld_root_env))
            elif _bld_root_rc:
                self._croot = abspath(expanduser(expandvars(_bld_root_rc)))
            elif root_writable:
                self._croot = join(root_dir, 'conda-bld')
            else:
                self._croot = abspath(expanduser('~/conda-bld'))
        return self._croot

    @croot.setter
    def croot(self, croot):
        """Set croot - if None is passed, then the default value will be used"""
        self._croot = croot

    @property
    def output_folder(self):
        return self._output_folder or self.croot

    @output_folder.setter
    def output_folder(self, value):
        self._output_folder = value

    @property
    def build_folder(self):
        """This is the core folder for a given build.
        It has the environments and work directories."""
        return os.path.join(self.croot, self.build_id)

    # back compat for conda-build-all - expects CONDA_* vars to be attributes of the config object
    @property
    @print_function_deprecation_warning
    def CONDA_LUA(self):
        return self.variant.get('lua', get_default_variant(self)['lua'])

    @CONDA_LUA.setter
    @print_function_deprecation_warning
    def CONDA_LUA(self, value):
        self.variant['lua'] = value

    @property
    @print_function_deprecation_warning
    def CONDA_PY(self):
        value = self.variant.get('python', get_default_variant(self)['python'])
        return int(''.join(value.split('.')))

    @CONDA_PY.setter
    @print_function_deprecation_warning
    def CONDA_PY(self, value):
        value = str(value)
        self.variant['python'] = '.'.join((value[0], value[1:]))

    @property
    @print_function_deprecation_warning
    def CONDA_NPY(self):
        value = self.variant.get('numpy', get_default_variant(self)['numpy'])
        return int(''.join(value.split('.')))

    @CONDA_NPY.setter
    @print_function_deprecation_warning
    def CONDA_NPY(self, value):
        value = str(value)
        self.variant['numpy'] = '.'.join((value[0], value[1:]))

    @property
    @print_function_deprecation_warning
    def CONDA_PERL(self):
        return self.variant.get('perl', get_default_variant(self)['perl'])

    @CONDA_PERL.setter
    @print_function_deprecation_warning
    def CONDA_PERL(self, value):
        self.variant['perl'] = value

    @property
    @print_function_deprecation_warning
    def CONDA_R(self):

        return self.variant.get('r_base', get_default_variant(self)['r_base'])

    @CONDA_R.setter
    @print_function_deprecation_warning
    def CONDA_R(self, value):
        self.variant['r_base'] = value

    def _get_python(self, prefix, platform):
        if platform.startswith('win'):
            if os.path.isfile(os.path.join(prefix, 'python_d.exe')):
                res = join(prefix, 'python_d.exe')
            else:
                res = join(prefix, 'python.exe')
        else:
            res = join(prefix, 'bin/python')
        return res

    def _get_perl(self, prefix, platform):
        if platform.startswith('win'):
            res = join(prefix, 'Library', 'bin', 'perl.exe')
        else:
            res = join(prefix, 'bin/perl')
        return res

    # TODO: This is probably broken on Windows, but no one has a lua package on windows to test.
    def _get_lua(self, prefix, platform):
        lua_ver = self.variant.get('lua', get_default_variant(self)['lua'])
        binary_name = "luajit" if (lua_ver and lua_ver[0] == "2") else "lua"
        if platform.startswith('win'):
            res = join(prefix, 'Library', 'bin', '{}.exe'.format(binary_name))
        else:
            res = join(prefix, 'bin/{}'.format(binary_name))
        return res

    def _get_r(self, prefix, platform):
        if platform.startswith('win'):
            res = join(prefix, 'Scripts', 'R.exe')
            # MRO test:
            if not os.path.exists(res):
                res = join(prefix, 'bin', 'R.exe')
        else:
            res = join(prefix, 'bin', 'R')
        return res

    def compute_build_id(self, package_name, reset=False):
        if self.set_build_id and (not self._build_id or reset):
            assert not os.path.isabs(package_name), ("package name should not be a absolute path, "
                                                     "to preserve croot during path joins")
            folder_basenames = [os.path.basename(fldr) for fldr in get_build_folders(self.croot)]
            build_folders = sorted([build_folder for build_folder in folder_basenames
                            if build_folder[:build_folder.rfind('_')] == package_name])
            if self.dirty and build_folders:
                # Use the most recent build with matching recipe name
                self._build_id = build_folders[-1]
            else:
                old_dir = self.work_dir if len(os.listdir(self.work_dir)) > 0 else ""
                # here we uniquely name folders, so that more than one build can happen concurrently
                #    keep 6 decimal places so that prefix < 80 chars
                build_id = package_name + "_" + str(int(time.time() * 1000))
                # important: this is recomputing prefixes and determines where work folders are.
                self._build_id = build_id
                if old_dir:
                    shutil.move(old_dir, self.work_dir)

    @property
    def build_id(self):
        """This is a per-build (almost) unique id, consisting of the package being built, and the
        time since the epoch, in ms.  It is appended to build and test prefixes, and used to create
        unique work folders for build and test."""
        return self._build_id

    @build_id.setter
    def build_id(self, _build_id):
        _build_id = _build_id.rstrip("/").rstrip("\\")
        assert not os.path.isabs(_build_id), ("build_id should not be an absolute path, "
                                              "to preserve croot during path joins")
        self._build_id = _build_id

    @property
    def prefix_length(self):
        return self._prefix_length

    @prefix_length.setter
    def prefix_length(self, length):
        self._prefix_length = length

    @property
    def _short_host_prefix(self):
        return join(self.build_folder, '_h_env')

    @property
    def _long_host_prefix(self):
        placeholder_length = self.prefix_length - len(self._short_host_prefix)
        placeholder = '_placehold'
        repeats = int(math.ceil(placeholder_length / len(placeholder)) + 1)
        placeholder = (self._short_host_prefix + repeats * placeholder)[:self.prefix_length]
        return max(self._short_host_prefix, placeholder)

    @property
    def build_prefix(self):
        """The temporary folder where the build environment is created.  The build env contains
        libraries that may be linked, but only if the host env is not specified.  It is placed on
        PATH."""
        if self.build_is_host:
            prefix = self.host_prefix
        else:
            prefix = join(self.build_folder, '_build_env')
        return prefix

    @property
    def host_prefix(self):
        """The temporary folder where the host environment is created.  The host env contains
        libraries that may be linked.  It is not placed on PATH."""
        if on_win:
            return self._short_host_prefix
        return self._long_host_prefix

    @property
    def _short_test_prefix(self):
        return join(self.build_folder, '_test_env')

    def _long_prefix(self, base_prefix):
        placeholder_length = self.prefix_length - len(base_prefix)
        placeholder = '_placehold'
        repeats = int(math.ceil(placeholder_length / len(placeholder)) + 1)
        placeholder = (base_prefix + repeats * placeholder)[:self.prefix_length]
        return max(base_prefix, placeholder)

    @property
    def test_prefix(self):
        """The temporary folder where the test environment is created"""
        if on_win or not self.long_test_prefix:
            return self._short_test_prefix
        return self._long_prefix(self._short_test_prefix)

    @property
    def build_python(self):
        return self.python_bin(self.build_prefix, self.platform)

    @property
    def host_python(self):
        return self._get_python(self.host_prefix, self.host_platform)

    @property
    def test_python(self):
        return self.python_bin(self.test_prefix, self.host_platform)

    def python_bin(self, prefix, platform):
        return self._get_python(prefix, platform)

    def perl_bin(self, prefix, platform):
        return self._get_perl(prefix, platform)

    def lua_bin(self, prefix, platform):
        return self._get_lua(prefix, platform)

    def r_bin(self, prefix, platform):
        return self._get_r(prefix, platform)

    @property
    def info_dir(self):
        """Path to the info dir in the build prefix, where recipe metadata is stored"""
        path = join(self.host_prefix, 'info')
        _ensure_dir(path)
        return path

    @property
    def meta_dir(self):
        """Path to the conda-meta dir in the build prefix, where package index json files are
        stored"""
        path = join(self.host_prefix, 'conda-meta')
        _ensure_dir(path)
        return path

    @property
    def broken_dir(self):
        """Where packages that fail the test phase are placed"""
        path = join(self.croot, "broken")
        _ensure_dir(path)
        return path

    @property
    def bldpkgs_dir(self):
        """ Dir where the package is saved. """
        path = join(self.croot, self.host_subdir)
        _ensure_dir(path)
        return path

    @property
    def bldpkgs_dirs(self):
        """ Dirs where previous build packages might be. """
        # The first two *might* be the same, but might not, depending on if this is a cross-compile.
        #     subdir should be the native platform, while self.subdir would be the host platform.
        return {join(self.croot, self.host_subdir), join(self.croot, subdir),
                join(self.croot, "noarch"), }

    @property
    def src_cache(self):
        """Where tarballs and zip files are downloaded and stored"""
        path = join(self.src_cache_root, 'src_cache')
        _ensure_dir(path)
        return path

    @property
    def git_cache(self):
        """Where local clones of git sources are stored"""
        path = join(self.src_cache_root, 'git_cache')
        _ensure_dir(path)
        return path

    @property
    def hg_cache(self):
        """Where local clones of hg sources are stored"""
        path = join(self.src_cache_root, 'hg_cache')
        _ensure_dir(path)
        return path

    @property
    def svn_cache(self):
        """Where local checkouts of svn sources are stored"""
        path = join(self.src_cache_root, 'svn_cache')
        _ensure_dir(path)
        return path

    @property
    def work_dir(self):
        """Where the source for the build is extracted/copied to."""
        path = join(self.build_folder, 'work')
        _ensure_dir(path)
        # if os.path.isdir(path):
        #     lst = [fn for fn in os.listdir(path) if not fn.startswith('.')]
        #     if len(lst) == 1:
        #         dir_path = join(path, lst[0])
        #         if os.path.isdir(dir_path):
        #             return dir_path
        return path

    @property
    def test_dir(self):
        """The temporary folder where test files are copied to, and where tests start execution"""
        path = join(self.build_folder, 'test_tmp')
        _ensure_dir(path)
        return path

    def clean(self, remove_folders=True):
        # build folder is the whole burrito containing envs and source folders
        #   It will only exist if we download source, or create a build or test environment
        if remove_folders and not getattr(self, 'dirty'):
            if self.build_id:
                if os.path.isdir(self.build_folder):
                    rm_rf(self.build_folder)
            else:
                for path in [self.work_dir, self.test_dir, self.build_prefix, self.test_prefix]:
                    if os.path.isdir(path):
                        rm_rf(path)
            if os.path.isfile(os.path.join(self.build_folder, 'prefix_files')):
                rm_rf(os.path.join(self.build_folder, 'prefix_files'))
        else:
            print("\nLeaving build/test directories:"
                  "\n  Work:\t", self.work_dir,
                  "\n  Test:\t", self.test_dir,
                  "\nLeaving build/test environments:"
                  "\n  Test:\tsource activate ", self.test_prefix,
                  "\n  Build:\tsource activate ", self.build_prefix,
                  "\n\n")

        for lock in get_conda_operation_locks(self.locking, self.bldpkgs_dirs):
            if os.path.isfile(lock.lock_file):
                rm_rf(lock.lock_file)

    def clean_pkgs(self):
        for folder in self.bldpkgs_dirs:
            rm_rf(folder)

    def copy(self):
        new = copy.copy(self)
        new.variant = copy.deepcopy(self.variant)
        if hasattr(self, 'variants'):
            new.variants = copy.deepcopy(self.variants)
        return new

    # context management - automatic cleanup if self.dirty or self.keep_old_work is not True
    def __enter__(self):
        pass

    def __exit__(self, e_type, e_value, traceback):
        if not getattr(self, 'dirty') and e_type is None and not getattr(self, 'keep_old_work'):
            get_logger(__name__).info("--dirty flag and --keep-old-work not specified."
                                        "Removing build/test folder after successful build/test.\n")
            self.clean()
        else:
            self.clean(remove_folders=False)


def get_or_merge_config(config, variant=None, **kwargs):
    """Always returns a new object - never changes the config that might be passed in."""
    if not config:
        config = Config(variant=variant)
    else:
        # decouple this config from whatever was fed in.  People must change config by
        #    accessing and changing this attribute.
        config = config.copy()
    if kwargs:
        config.set_keys(**kwargs)
    if variant:
        config.variant.update(variant)
    return config


# legacy exports for conda
croot = Config().croot



# (c) Continuum Analytics, Inc. / http://continuum.io
# All Rights Reserved
#
# conda is distributed under the terms of the BSD 3-clause license.
# Consult LICENSE.txt or http://opensource.org/licenses/BSD-3-Clause.

from __future__ import absolute_import, division, print_function

from collections import OrderedDict
from locale import getpreferredencoding
import os
from os.path import isdir, isfile, abspath
import random
import re
import subprocess
import string
import sys
import tarfile
import tempfile

import yaml

from .conda_interface import (PY3, UnsatisfiableError, ProgressiveFetchExtract,
                              TemporaryDirectory)
from .conda_interface import execute_actions
from .conda_interface import pkgs_dirs
from .conda_interface import conda_43
from .conda_interface import specs_from_url
from .conda_interface import memoized

from conda_build import exceptions, utils, environ
from conda_build.metadata import MetaData, combine_top_level_metadata_with_output
import conda_build.source as source
from conda_build.variants import (get_package_variants, list_of_dicts_to_dict_of_lists,
                                  filter_by_key_value)
from conda_build.exceptions import DependencyNeedsBuildingError
from conda_build.index import get_build_index
# from conda_build.jinja_context import pin_subpackage_against_outputs

yaml.add_representer(set, yaml.representer.SafeRepresenter.represent_list)
yaml.add_representer(tuple, yaml.representer.SafeRepresenter.represent_list)


def bldpkg_path(m):
    '''
    Returns path to built package's tarball given its ``Metadata``.
    '''
    subdir = 'noarch' if m.noarch or m.noarch_python else m.config.host_subdir
    return os.path.join(m.config.output_folder, subdir, '%s.tar.bz2' % m.dist())


def actions_to_pins(actions):
    specs = []
    if conda_43:
        spec_name = lambda x: x.dist_name
    else:
        spec_name = lambda x: x
    if 'LINK' in actions:
        specs = [' '.join(spec_name(spec).split()[0].rsplit('-', 2)) for spec in actions['LINK']]
    return specs


def get_env_dependencies(m, env, variant, exclude_pattern=None,
                         permit_unsatisfiable_variants=False,
                         merge_build_host_on_same_platform=True):
    dash_or_under = re.compile("[-_]")
    specs = m.get_depends_top_and_out(env)
    # replace x.x with our variant's numpy version, or else conda tries to literally go get x.x
    if env in ('build', 'host'):
        no_xx_specs = []
        for spec in specs:
            if ' x.x' in spec:
                pkg_name = spec.split()[0]
                no_xx_specs.append(' '.join((pkg_name, variant.get(pkg_name, ""))))
            else:
                no_xx_specs.append(spec)
        specs = no_xx_specs
    subpackages = []
    dependencies = []
    pass_through_deps = []
    # ones that get filtered from actual versioning, to exclude them from the hash calculation
    for spec in specs:
        if not exclude_pattern or not exclude_pattern.match(spec):
            is_subpackage = False
            spec_name = spec.split()[0]
            for entry in m.get_section('outputs'):
                name = entry.get('name')
                if name == spec_name:
                    subpackages.append(' '.join((name, m.version())))
                    is_subpackage = True
            if not is_subpackage:
                dependencies.append(spec)
            # fill in variant version iff no version at all is provided
            for key, value in variant.items():
                # for sake of comparison, ignore dashes and underscores
                if (dash_or_under.sub("", key) == dash_or_under.sub("", spec_name) and
                        not re.search(r'%s\s+[0-9a-zA-Z\_\.\<\>\=\*]' % spec_name, spec)):
                    dependencies.append(" ".join((spec_name, value)))
        elif exclude_pattern.match(spec):
            pass_through_deps.append(spec)
    random_string = ''.join(random.choice(string.ascii_uppercase + string.digits)
                            for _ in range(10))
    dependencies = set(dependencies)
    unsat = None
    with TemporaryDirectory(prefix="_", suffix=random_string) as tmpdir:
        try:
            actions = environ.get_install_actions(tmpdir, tuple(dependencies), env,
                                                  subdir=getattr(m.config, '{}_subdir'.format(env)),
                                                  debug=m.config.debug,
                                                  verbose=m.config.verbose,
                                                  locking=m.config.locking,
                                                  bldpkgs_dirs=tuple(m.config.bldpkgs_dirs),
                                                  timeout=m.config.timeout,
                                                  disable_pip=m.config.disable_pip,
                                                  max_env_retry=m.config.max_env_retry,
                                                  output_folder=m.config.output_folder,
                                                  channel_urls=tuple(m.config.channel_urls))
        except (UnsatisfiableError, DependencyNeedsBuildingError) as e:
            # we'll get here if the environment is unsatisfiable
            if hasattr(e, 'packages'):
                unsat = ', '.join(e.packages)
            else:
                unsat = e.message
            if permit_unsatisfiable_variants:
                actions = {}
            else:
                raise

    specs = actions_to_pins(actions)
    return ((specs + subpackages + pass_through_deps) or
                  m.meta.get('requirements', {}).get(env, []),
            actions, unsat)


def strip_channel(spec_str):
    if hasattr(spec_str, 'decode'):
        spec_str = spec_str.decode()
    if ':' in spec_str:
        spec_str = spec_str.split("::")[-1]
    return spec_str


def get_pin_from_build(m, dep, build_dep_versions):
    dep_name = dep.split()[0]
    pin = None
    version = build_dep_versions.get(dep_name) or m.config.variant.get(dep_name)
    if (version and dep_name in m.config.variant.get('pin_run_as_build', {}) and
            not (dep_name == 'python' and (m.noarch or m.noarch_python)) and
            dep_name in build_dep_versions):
        pin_cfg = m.config.variant['pin_run_as_build'][dep_name]
        if isinstance(pin_cfg, str):
            # if pin arg is a single 'x.x', use the same value for min and max
            pin_cfg = dict(min_pin=pin_cfg, max_pin=pin_cfg)
        pin = utils.apply_pin_expressions(version.split()[0], **pin_cfg)
    elif dep.startswith('numpy') and 'x.x' in dep:
        if not build_dep_versions.get(dep_name):
            raise ValueError("numpy x.x specified, but numpy not in build requirements.")
        pin = utils.apply_pin_expressions(version.split()[0], min_pin='x.x', max_pin='x.x')
    if pin:
        dep = " ".join((dep_name, pin))
    return dep


def _filter_run_exports(specs, ignore_list):
    filtered_specs = {}
    for agent, specs_list in specs.items():
        for spec in specs_list:
            if hasattr(spec, 'decode'):
                spec = spec.decode()
            if not any((spec == ignore_spec or spec.startswith(ignore_spec + ' '))
                       for ignore_spec in ignore_list):
                filtered_specs[agent] = filtered_specs.get(agent, []) + [spec]
    return filtered_specs


def find_pkg_dir_or_file_in_pkgs_dirs(pkg_dist, m):
    _pkgs_dirs = pkgs_dirs + list(m.config.bldpkgs_dirs)
    pkg_loc = None
    for pkgs_dir in _pkgs_dirs:
        pkg_dir = os.path.join(pkgs_dir, pkg_dist)
        pkg_file = os.path.join(pkgs_dir, pkg_dist + '.tar.bz2')
        if os.path.isdir(pkg_dir):
            pkg_loc = pkg_dir
            break
        elif os.path.isfile(pkg_file):
            pkg_loc = pkg_file
            break
    return pkg_loc


@memoized
def _read_specs_from_package(pkg_loc, pkg_dist):
    specs = {}
    if os.path.isdir(pkg_loc):
        downstream_file = os.path.join(pkg_loc, 'info/run_exports')
        if os.path.isfile(downstream_file):
            with open(downstream_file) as f:
                specs = {'weak': [spec.rstrip() for spec in f.readlines()]}
        # a later attempt: record more info in the yaml file, to support "strong" run exports
        elif os.path.isfile(downstream_file + '.yaml'):
            with open(downstream_file + '.yaml') as f:
                specs = yaml.safe_load(f)
    if not specs and os.path.isfile(pkg_loc):
        specs_yaml = utils.package_has_file(pkg_loc, 'info/run_exports.yaml')
        if specs_yaml:
            specs = yaml.safe_load(specs_yaml)
        else:
            legacy_specs = utils.package_has_file(pkg_loc, 'info/run_exports')
            # exclude packages pinning themselves (makes no sense)
            if legacy_specs:
                specs = {'weak': [spec.rstrip() for spec in legacy_specs.splitlines()
                                if not spec.startswith(pkg_dist.rsplit('-', 2)[0])]}
    return specs


def get_upstream_pins(m, actions, env):
    """Download packages from specs, then inspect each downloaded package for additional
    downstream dependency specs.  Return these additional specs."""

    env_specs = m.meta.get('requirements', {}).get(env, [])
    explicit_specs = [req.split(' ')[0] for req in env_specs] if env_specs else []
    linked_packages = actions.get('LINK', [])
    linked_packages = [pkg for pkg in linked_packages if pkg.name in explicit_specs]

    # edit the plan to download all necessary packages
    for key in ('LINK', 'EXTRACT', 'UNLINK'):
        if key in actions:
            del actions[key]
    # this should be just downloading packages.  We don't need to extract them -
    #    we read contents directly

    index, index_ts = get_build_index(getattr(m.config, '{}_subdir'.format(env)),
                                      bldpkgs_dir=m.config.bldpkgs_dir,
                                      output_folder=m.config.output_folder,
                                      channel_urls=m.config.channel_urls,
                                      debug=m.config.debug, verbose=m.config.verbose,
                                      locking=m.config.locking, timeout=m.config.timeout)
    if 'FETCH' in actions or 'EXTRACT' in actions:
        # this is to force the download
        execute_actions(actions, index, verbose=m.config.debug)
    ignore_list = utils.ensure_list(m.get_value('build/ignore_run_exports'))

    additional_specs = {}
    for pkg in linked_packages:
        if hasattr(pkg, 'dist_name'):
            pkg_dist = pkg.dist_name
        else:
            pkg = strip_channel(pkg)
            pkg_dist = pkg.split(' ')[0]
        pkg_loc = find_pkg_dir_or_file_in_pkgs_dirs(pkg_dist, m)

        # ran through all pkgs_dirs, and did not find package or folder.  Download it.
        # TODO: this is a vile hack reaching into conda's internals. Replace with
        #    proper conda API when available.
        if not pkg_loc and conda_43:
            try:
                # the conda 4.4 API uses a single `link_prefs` kwarg
                # whereas conda 4.3 used `index` and `link_dists` kwargs
                pfe = ProgressiveFetchExtract(link_prefs=(index[pkg],))
            except TypeError:
                # TypeError: __init__() got an unexpected keyword argument 'link_prefs'
                pfe = ProgressiveFetchExtract(link_dists=[pkg], index=index)
            with utils.LoggingContext():
                pfe.execute()
            for pkg_dir in pkgs_dirs:
                _loc = os.path.join(pkg_dir, index[pkg].fn)
                if os.path.isfile(_loc):
                    pkg_loc = _loc
                    break

        specs = _read_specs_from_package(pkg_loc, pkg_dist)

        additional_specs = utils.merge_dicts_of_lists(additional_specs,
                                                      _filter_run_exports(specs, ignore_list))
    return additional_specs


def _read_upstream_pin_files(m, env, permit_unsatisfiable_variants, exclude_pattern):
    deps, actions, unsat = get_env_dependencies(m, env, m.config.variant,
                                exclude_pattern,
                                permit_unsatisfiable_variants=permit_unsatisfiable_variants)
    # extend host deps with strong build run exports.  This is important for things like
    #    vc feature activation to work correctly in the host env.
    extra_run_specs = get_upstream_pins(m, actions, env)
    return list(set(deps)) or m.meta.get('requirements', {}).get(env, []), unsat, extra_run_specs


def add_upstream_pins(m, permit_unsatisfiable_variants, exclude_pattern):
    """Applies run_exports from any build deps to host and run sections"""
    # if we have host deps, they're more important than the build deps.
    requirements = m.meta.get('requirements', {})
    build_deps, build_unsat, extra_run_specs_from_build = _read_upstream_pin_files(m, 'build',
                                                    permit_unsatisfiable_variants, exclude_pattern)

    # is there a 'host' section?
    if m.is_cross:
        # this must come before we read upstream pins, because it will enforce things
        #      like vc version from the compiler.
        if m.meta.get('requirements', {}).get('host'):
            host_reqs = m.meta['requirements']['host']
        else:
            matching_output = [out for out in m.meta['outputs'] if out.get('name') == m.name()]
            if matching_output:
                requirements = utils.expand_reqs(matching_output[0]['requirements'])
                matching_output[0]['requirements'] = requirements
                host_reqs = requirements['host']
        # in-place modification of above thingie
        host_reqs.extend(extra_run_specs_from_build.get('strong', []))

        host_deps, host_unsat, extra_run_specs_from_host = _read_upstream_pin_files(m, 'host',
                                                    permit_unsatisfiable_variants, exclude_pattern)
        extra_run_specs = set(extra_run_specs_from_host.get('strong', []) +
                              extra_run_specs_from_host.get('weak', []) +
                              extra_run_specs_from_build.get('strong', []))
    else:
        host_deps = []
        host_unsat = []
        extra_run_specs = set(extra_run_specs_from_build.get('strong', []))
        if not m.uses_new_style_compiler_activation:
            extra_run_specs.update(extra_run_specs_from_build.get('weak', []))

    run_deps = extra_run_specs | set(utils.ensure_list(requirements.get('run')))

    for section, deps in (('build', build_deps), ('host', host_deps), ('run', run_deps)):
        if deps:
            requirements[section] = list(deps)

    m.meta['requirements'] = requirements
    return build_unsat, host_unsat


def finalize_metadata(m, permit_unsatisfiable_variants=False):
    """Fully render a recipe.  Fill in versions for build/host dependencies."""
    if m.skip():
        rendered_metadata = m.copy()
        rendered_metadata.final = True
    else:
        exclude_pattern = None
        excludes = set(m.config.variant.get('ignore_version', []))

        for key in m.config.variant.get('pin_run_as_build', {}).keys():
            if key in excludes:
                excludes.remove(key)

        output_excludes = set()
        if hasattr(m, 'other_outputs'):
            output_excludes = set(name for (name, variant) in m.other_outputs.keys())

        if excludes or output_excludes:
            exclude_pattern = re.compile('|'.join('(?:^{}(?:\s|$|\Z))'.format(exc)
                                            for exc in excludes | output_excludes))

        parent_recipe = m.meta.get('extra', {}).get('parent_recipe', {})

        # extract the topmost section where variables are defined, and put it on top of the
        #     requirements for a particular output
        # Re-parse the output from the original recipe, so that we re-consider any jinja2 stuff
        output = m.get_rendered_output(m.name())
        rendered_metadata = m.get_output_metadata(output)

        if output:
            if 'package' in output or 'name' not in output:
                # it's just a top-level recipe
                output = {'name': m.name()}

            if not parent_recipe or parent_recipe['name'] == m.name():
                combine_top_level_metadata_with_output(rendered_metadata, output)
            requirements = utils.expand_reqs(output.get('requirements', {}))
            rendered_metadata.meta['requirements'] = requirements

        if rendered_metadata.meta.get('requirements'):
            utils.insert_variant_versions(rendered_metadata.meta['requirements'],
                                          rendered_metadata.config.variant, 'build')
            utils.insert_variant_versions(rendered_metadata.meta['requirements'],
                                        rendered_metadata.config.variant, 'host')

        build_unsat, host_unsat = add_upstream_pins(rendered_metadata,
                                                    permit_unsatisfiable_variants,
                                                    exclude_pattern)
        # getting this AFTER add_upstream_pins is important, because that function adds deps
        #     to the metadata.
        requirements = rendered_metadata.meta.get('requirements', {})

        # here's where we pin run dependencies to their build time versions.  This happens based
        #     on the keys in the 'pin_run_as_build' key in the variant, which is a list of package
        #     names to have this behavior.
        if output_excludes:
            exclude_pattern = re.compile('|'.join('(?:^{}(?:\s|$|\Z))'.format(exc)
                                            for exc in output_excludes))
        pinning_env = 'host' if rendered_metadata.is_cross else 'build'

        build_reqs = requirements.get(pinning_env, [])
        # if python is in the build specs, but doesn't have a specific associated
        #    version, make sure to add one
        if build_reqs and 'python' in build_reqs:
            build_reqs.append('python {}'.format(m.config.variant['python']))
            rendered_metadata.meta['requirements'][pinning_env] = build_reqs

        full_build_deps, _, _ = get_env_dependencies(rendered_metadata, pinning_env,
                                        rendered_metadata.config.variant,
                                        exclude_pattern=exclude_pattern,
                                        permit_unsatisfiable_variants=permit_unsatisfiable_variants)
        full_build_dep_versions = {dep.split()[0]: " ".join(dep.split()[1:])
                                   for dep in full_build_deps}

        if isfile(m.requirements_path) and not requirements.get('run'):
            requirements['run'] = specs_from_url(m.requirements_path)
        run_deps = requirements.get('run', [])

        versioned_run_deps = [get_pin_from_build(rendered_metadata, dep, full_build_dep_versions)
                            for dep in run_deps]
        versioned_run_deps = [utils.ensure_valid_spec(spec, warn=True)
                              for spec in versioned_run_deps]
        requirements[pinning_env] = full_build_deps
        requirements['run'] = versioned_run_deps

        rendered_metadata.meta['requirements'] = requirements

        # append other requirements, such as python.app, appropriately
        rendered_metadata.append_requirements()

        if rendered_metadata.pin_depends == 'strict':
            rendered_metadata.meta['requirements']['run'] = environ.get_pinned_deps(
                rendered_metadata, 'run')
        test_deps = rendered_metadata.get_value('test/requires')
        if test_deps:
            versioned_test_deps = list({get_pin_from_build(m, dep, full_build_dep_versions)
                                        for dep in test_deps})
            versioned_test_deps = [utils.ensure_valid_spec(spec, warn=True)
                                for spec in versioned_test_deps]
            rendered_metadata.meta['test']['requires'] = versioned_test_deps
        extra = rendered_metadata.meta.get('extra', {})
        extra['copy_test_source_files'] = m.config.copy_test_source_files
        rendered_metadata.meta['extra'] = extra

        # if source/path is relative, then the output package makes no sense at all.  The next
        #   best thing is to hard-code the absolute path.  This probably won't exist on any
        #   system other than the original build machine, but at least it will work there.
        if m.meta.get('source'):
            if 'path' in m.meta['source'] and not os.path.isabs(m.meta['source']['path']):
                rendered_metadata.meta['source']['path'] = os.path.normpath(
                    os.path.join(m.path, m.meta['source']['path']))
            elif ('git_url' in m.meta['source'] and not (
                    # absolute paths are not relative paths
                    os.path.isabs(m.meta['source']['git_url']) or
                    # real urls are not relative paths
                    ":" in m.meta['source']['git_url'])):
                rendered_metadata.meta['source']['git_url'] = os.path.normpath(
                    os.path.join(m.path, m.meta['source']['git_url']))

        if not rendered_metadata.meta.get('build'):
            rendered_metadata.meta['build'] = {}

        if build_unsat or host_unsat:
            rendered_metadata.final = False
            log = utils.get_logger(__name__)
            log.warn("Returning non-final recipe for {}; one or more dependencies "
                    "was unsatisfiable:\nBuild: {}\nHost: {}".format(rendered_metadata.dist(),
                                                                    build_unsat, host_unsat))
        else:
            rendered_metadata.final = True
    return rendered_metadata


def try_download(metadata, no_download_source, raise_error=False):
    if not metadata.source_provided and not no_download_source:
        # this try/catch is for when the tool to download source is actually in
        #    meta.yaml, and not previously installed in builder env.
        try:
            source.provide(metadata)
        except subprocess.CalledProcessError as error:
            print("Warning: failed to download source.  If building, will try "
                "again after downloading recipe dependencies.")
            print("Error was: ")
            print(error)

    if not metadata.source_provided:
        if no_download_source:
            raise ValueError("no_download_source specified, but can't fully render recipe without"
                             " downloading source.  Please fix the recipe, or don't use "
                             "no_download_source.")
        elif raise_error:
            raise RuntimeError("Failed to download or patch source. Please see build log for info.")


def reparse(metadata):
    """Some things need to be parsed again after the build environment has been created
    and activated."""
    metadata.final = False
    sys.path.insert(0, metadata.config.build_prefix)
    sys.path.insert(0, metadata.config.host_prefix)
    py_ver = '.'.join(metadata.config.variant['python'].split('.')[:2])
    sys.path.insert(0, utils.get_site_packages(metadata.config.host_prefix, py_ver))
    metadata.parse_until_resolved()
    metadata = finalize_metadata(metadata)
    return metadata


def distribute_variants(metadata, variants, permit_unsatisfiable_variants=False,
                        allow_no_other_outputs=False, bypass_env_check=False):
    rendered_metadata = {}
    need_source_download = True

    # don't bother distributing python if it's a noarch package
    if metadata.noarch or metadata.noarch_python:
        variants = filter_by_key_value(variants, 'python', variants[0]['python'],
                                       'noarch_reduction')

    # store these for reference later
    metadata.config.variants = variants
    # These are always the full set.  just 'variants' is the one that gets
    #     used mostly, and can be reduced
    metadata.config.input_variants = variants

    recipe_requirements = metadata.extract_requirements_text()
    recipe_package_and_build_text = metadata.extract_package_and_build_text()
    recipe_text = recipe_package_and_build_text + recipe_requirements
    if PY3 and hasattr(recipe_text, 'decode'):
        recipe_text = recipe_text.decode()
    elif not PY3 and hasattr(recipe_text, 'encode'):
        recipe_text = recipe_text.encode()

    for variant in variants:
        mv = metadata.copy()
        mv.config.variant = variant
        used_variables = mv.get_used_loop_vars()
        conform_dict = {}
        for key in used_variables:
            # We use this variant in the top-level recipe.
            # constrain the stored variants to only this version in the output
            #     variant mapping
            conform_dict[key] = variant[key]

        build_reqs = mv.meta.get('requirements', {}).get('build', [])
        host_reqs = mv.meta.get('requirements', {}).get('host', [])

        if 'python' in build_reqs or 'python' in host_reqs:
            conform_dict['python'] = variant['python']
        if 'r-base' in build_reqs or 'r-base' in host_reqs:
            conform_dict['r_base'] = variant['r_base']

        pin_run_as_build = variant.get('pin_run_as_build', {})
        if mv.numpy_xx and 'numpy' not in pin_run_as_build:
            pin_run_as_build['numpy'] = {'min_pin': 'x.x', 'max_pin': 'x.x'}

        for key, values in conform_dict.items():
            mv.config.variants = (filter_by_key_value(mv.config.variants, key, values,
                                                      'distribute_variants_reduction') or
                                  mv.config.variants)
        numpy_pinned_variants = []
        for _variant in mv.config.variants:
            _variant['pin_run_as_build'] = pin_run_as_build
            numpy_pinned_variants.append(_variant)
        mv.config.variants = numpy_pinned_variants

        mv.config.squished_variants = list_of_dicts_to_dict_of_lists(mv.config.variants)

        if mv.needs_source_for_render and mv.variant_in_source:
            mv.parse_again()
            utils.rm_rf(mv.config.work_dir)
            source.provide(mv)
            mv.parse_again()

        mv.parse_until_resolved(allow_no_other_outputs=allow_no_other_outputs,
                                bypass_env_check=bypass_env_check)
        need_source_download = (not mv.needs_source_for_render or not mv.source_provided)

        rendered_metadata[(mv.dist(),
                           mv.config.variant.get('target_platform', mv.config.subdir),
                           tuple((var, mv.config.variant[var])
                                 for var in mv.get_used_vars()))] = \
                                    (mv, need_source_download, None)
    # list of tuples.
    # each tuple item is a tuple of 3 items:
    #    metadata, need_download, need_reparse_in_env
    return list(rendered_metadata.values())


def expand_outputs(metadata_tuples):
    """Obtain all metadata objects for all outputs from recipe.  Useful for outputting paths."""
    expanded_outputs = OrderedDict()
    for (_m, download, reparse) in metadata_tuples:
        for (output_dict, m) in _m.get_output_metadata_set(permit_unsatisfiable_variants=False):
            expanded_outputs[m.dist()] = (output_dict, m)
    return list(expanded_outputs.values())


def render_recipe(recipe_path, config, no_download_source=False, variants=None,
                  permit_unsatisfiable_variants=True, reset_build_id=True, bypass_env_check=False):
    """Returns a list of tuples, each consisting of

    (metadata-object, needs_download, needs_render_in_env)

    You get one tuple per variant.  Outputs are not factored in here (subpackages won't affect these
    results returned here.)
    """
    arg = recipe_path
    # Don't use byte literals for paths in Python 2
    if not PY3:
        arg = arg.decode(getpreferredencoding() or 'utf-8')
    if isfile(arg):
        if arg.endswith(('.tar', '.tar.gz', '.tgz', '.tar.bz2')):
            recipe_dir = tempfile.mkdtemp()
            t = tarfile.open(arg, 'r:*')
            t.extractall(path=recipe_dir)
            t.close()
            need_cleanup = True
        elif arg.endswith('.yaml'):
            recipe_dir = os.path.dirname(arg)
            need_cleanup = False
        else:
            print("Ignoring non-recipe: %s" % arg)
            return None, None
    else:
        recipe_dir = abspath(arg)
        need_cleanup = False

    if not isdir(recipe_dir):
        sys.exit("Error: no such directory: %s" % recipe_dir)

    try:
        m = MetaData(recipe_dir, config=config)
    except exceptions.YamlParsingError as e:
        sys.stderr.write(e.error_msg())
        sys.exit(1)

    rendered_metadata = {}

    # important: set build id *before* downloading source.  Otherwise source goes into a different
    #    build folder.
    if config.set_build_id:
        m.config.compute_build_id(m.name(), reset=reset_build_id)

    # this source may go into a folder that doesn't match the eventual build folder.
    #   There's no way around it AFAICT.  We must download the source to be able to render
    #   the recipe (from anything like GIT_FULL_HASH), but we can't know the final build
    #   folder until rendering is complete, because package names can have variant jinja2 in them.
    if m.needs_source_for_render and not m.source_provided:
        try_download(m, no_download_source=no_download_source)

    if m.final:
        if not hasattr(m.config, 'variants') or not m.config.variant:
            m.config.ignore_system_variants = True
            if os.path.isfile(os.path.join(m.path, 'conda_build_config.yaml')):
                m.config.variant_config_files = [os.path.join(m.path, 'conda_build_config.yaml')]
            m.config.variants = get_package_variants(m, variants=variants)
            m.config.variant = m.config.variants[0]
        rendered_metadata = [(m, False, False), ]
    else:
        index, index_ts = get_build_index(m.config.build_subdir,
                                          bldpkgs_dir=m.config.bldpkgs_dir,
                                          output_folder=m.config.output_folder,
                                          channel_urls=m.config.channel_urls,
                                          omit_defaults=m.config.override_channels,
                                          debug=m.config.debug, verbose=m.config.verbose,
                                          locking=m.config.locking, timeout=m.config.timeout)

        # merge any passed-in variants with any files found
        variants = get_package_variants(m, variants=variants)

        # when building, we don't want to fully expand all outputs into metadata, only expand
        #    whatever variants we have (i.e. expand top-level variants, not output-only variants)
        rendered_metadata = distribute_variants(m, variants,
                                    permit_unsatisfiable_variants=permit_unsatisfiable_variants,
                                    allow_no_other_outputs=True, bypass_env_check=bypass_env_check)

    if need_cleanup:
        utils.rm_rf(recipe_dir)

    return rendered_metadata


# Keep this out of the function below so it can be imported by other modules.
FIELDS = ["package", "source", "build", "requirements", "test", "app", "outputs", "about", "extra"]


# Next bit of stuff is to support YAML output in the order we expect.
# http://stackoverflow.com/a/17310199/1170370
class _MetaYaml(dict):
    fields = FIELDS

    def to_omap(self):
        return [(field, self[field]) for field in _MetaYaml.fields if field in self]


def _represent_omap(dumper, data):
    return dumper.represent_mapping(u'tag:yaml.org,2002:map', data.to_omap())


def _unicode_representer(dumper, uni):
    node = yaml.ScalarNode(tag=u'tag:yaml.org,2002:str', value=uni)
    return node


class _IndentDumper(yaml.Dumper):
    def increase_indent(self, flow=False, indentless=False):
        return super(_IndentDumper, self).increase_indent(flow, False)


yaml.add_representer(_MetaYaml, _represent_omap)
if PY3:
    yaml.add_representer(str, _unicode_representer)
    unicode = None  # silence pyflakes about unicode not existing in py3
else:
    yaml.add_representer(unicode, _unicode_representer)


def output_yaml(metadata, filename=None):
    utils.trim_empty_keys(metadata.meta)
    output = yaml.dump(_MetaYaml(metadata.meta), Dumper=_IndentDumper,
                       default_flow_style=False, indent=4)
    if filename:
        if any(sep in filename for sep in ('\\', '/')):
            try:
                os.makedirs(os.path.dirname(filename))
            except OSError:
                pass
        with open(filename, "w") as f:
            f.write(output)
        return "Wrote yaml to %s" % filename
    else:
        return output
